#!/usr/bin/env python
# -*- coding: utf-8 -*-

# # 2A.eco - Exercice API SNCF corrigé
# 
# Manipulation d'une [API REST](https://fr.wikipedia.org/wiki/Representational_state_transfer), celle de la SNCF est prise comme exemple. Correction d'exercices.

# In[1]:


from jyquickhelper import add_notebook_menu
add_notebook_menu()


# ## Partie 0 - modules recommandés et connexion à l'API 
# 
# Il vous faudra sûrement les modules suivant : 
# 
# - requests
# - datetime
# - pandas
# - matplotlib
# 
# Créer un login pour vous connecter à l'API de la SNCF https://data.sncf.com/api 
# 
# Vous pouvez maintenant commencer. Ce notebook peut prendre du temps à s'éxécuter, surout à partir de la partie 3

# In[2]:


# !!!!! Attention à bien mettre votre token ici  !!!!!

token_auth = 'XXXXXXXXXXXXXXXXXXXXXXXXXXXXXX'


# In[3]:


import os
from pyquickhelper.loghelper import get_password
if "XXXXXX" in token_auth:
    token_auth = get_password("sncf", "ensae_teaching_cs,key")


# ## Partie 1 - Trouver les gares accessibles _via_ la SNCF
# 
# - Trouver l'ensemble des gares disponibles sur l'API et créer un fichier csv avec les codes de la gare, son nom et ses coordonnées latitude et longitude, ainsi que les informations administratives de la région quand elles sont disponibles
# 
# - Représentez les sur un graphique

# In[4]:


import pandas as pd
import requests


# In[5]:


def page_gares(numero_page) :
    return requests.get(
        ('https://api.sncf.com/v1/coverage/sncf/stop_areas?start_page={}').format(numero_page),
        auth=(token_auth, ''))

######################################
# on commence par la première page qui nous donne le nombre de résultats par page ainsi que le nombre total de résultats                    

page_initiale = page_gares(0)                        
item_per_page = page_initiale.json()['pagination']['items_per_page']
total_items = page_initiale.json()['pagination']['total_result']
dfs = []

# on fait une boucle sur toutes les pages suivantes 
print_done = {}

for page in range(int(total_items/item_per_page)+1) : 
    stations_page = page_gares(page)
    
    ensemble_stations = stations_page.json()
    
    if 'stop_areas' not in ensemble_stations:
        # pas d'arrêt
        continue
    
    # on ne retient que les informations qui nous intéressent 
    for station in ensemble_stations['stop_areas']:

        station['lat'] = station['coord']['lat']
        station["lon"]  = station['coord']['lon']

        if 'administrative_regions' in station.keys() : 
            for var_api, var_df in zip(['insee','name','label','id','zip_code'],
                                       ['insee','region','label_region','id_region','zip_code']): 
                try:
                    station[var_df] = station['administrative_regions'][0][var_api]
                except KeyError:
                    if var_api not in print_done:
                        print("key '{0}' not here but {1}".format(var_api, 
                                ",".join(station['administrative_regions'][0].keys())))
                        print_done[var_api] = var_api

        [station.pop(k,None) for k in ['coord','links','administrative_regions', 'type', 'codes']]

    stations = ensemble_stations['stop_areas']
    try:
        dp = pd.DataFrame(stations)
    except Exception as e:
        # La SNCF modifie parfois le schéma de ses données.
        # On affiche station pour avoir une meilleure idée que l'erreur retournée par pandas
        raise Exception("Problème de données\n{0}".format(stations)) from e
        
    dfs.append(dp)
    if page % 10 == 0:
        print("je suis à la page", page, "---", dp.shape)

import pandas
df = pandas.concat(dfs)
df.to_csv("./ensemble_gares.csv")  
print(df.shape)
df.head()


# In[6]:


df[df.name.str.contains('Lyon')]


# In[7]:


df = pd.read_csv("./ensemble_gares.csv", encoding = "ISO-8859-1")
print(df.columns)
print(df.shape)
# Exemple des informations sur une gare
df.iloc[317]


# In[8]:


# on crée un dictionnaire des correspondances entre les noms et les codes des gares
dict_label_gare_code = df[['label','id']].set_index('label').to_dict()['id']
dict_nom_gare_code = df[['name','id']].set_index('name').to_dict()['id']


# In[9]:


print(df.columns)

# graphique dans le plan des gares
get_ipython().run_line_magic('matplotlib', 'inline')
import matplotlib.pyplot as plt
lng_var = df[(df['lat']>35) & (df['lat']<60)]["lon"].tolist()
lat_var = df[(df['lat']>35) & (df['lat']<60)]["lat"].tolist()
plt.scatter(x = lng_var , y = lat_var,marker = "o");


# ## Les trajets depuis la Gare de Lyon
# 
# ### Partons à Lyon
# 
# Imaginez que vous vouliez un peu voyager hors de Paris, et il se trouve que justement on vous propose de passer quelques jours à Lyon. Vous partez le 17 novembre vers 19h50 pour ne pas trop écourter votre journée de travail ([documentation de l'API](https://doc.navitia.io/#api-catalog)).

# #### Question 1
# 
# - Commencez par récupérer les informations sur le trajet entre Paris Gare de Lyon et Lyon Perrache le 17 novembre à 19h57
# 
#     - Paris - Gare de Lyon (code de la gare : __stop\_area:SNCF:87686006__)
# 
#     - Lyon - Gare Lyon Perrache (code de la gare : __stop\_area:SNCF:87722025__)
#     
#     - Indice : utiliser la requête "journeys"
#     
#     - Autre indice : le format de la date est AAAAMMJJTHHMMSS (Année, mois, jour, heure, minutes, secondes)
#     
# - Répondez aux questions suivantes 
#     - combien y a-t-il d'arrêts entre ces deux gares ? (utilisez la clé 'journeys')
#     - combien de temps d'arrêt à chacune d'elles ?

# Tout d'abord une fonction qui sera utile pour calculer des temps :

# In[10]:


from datetime import datetime, timedelta

def convertir_en_temps(chaine) : 
    ''' on convertit en date la chaine de caractères de l API'''
    return datetime.strptime(chaine.replace('T',''),'%Y%m%d%H%M%S')


# Et l'inverse :

# In[11]:


def convertir_en_chaine(dt) : 
    ''' on convertit en chaîne de caractères un datetime'''
    return datetime.strftime(dt, '%Y%m%dT%H%M%S')

now = datetime.now()
convertir_en_chaine(now)


# In[12]:


convertir_en_temps(convertir_en_chaine(now))


# In[13]:


# informations sur le trajet qu'on choisit dans le futur
# l'API ne retourne pas de résultatq très loin dans le passé
now = datetime.now()
dt = now + timedelta(14)  # dans deux semaines

date_depart = convertir_en_chaine(dt)
gare_depart = 'stop_area:SNCF:87686006'
gare_arrivee = 'stop_area:SNCF:87722025'

date_depart


# Ensemble des départs :

# In[14]:


paris_lyon = requests.get(
    ('https://api.sncf.com/v1/coverage/sncf/journeys?'
     'from={}&to={}&datetime={}').format(gare_depart, gare_arrivee, date_depart),
    auth=(token_auth, '')).json()


# Les gares du chemin entre Paris et Lyon sur ce trajet ainsi que le temps d'arrêt :

# In[15]:


session = paris_lyon['journeys'][0]['sections'][1]
rows = []
if "stop_date_times" in session:
    for i in session['stop_date_times']:
        rows.append(dict(name=i['stop_point']['name'],
                         depart=convertir_en_temps(i['departure_date_time']),
                         arrivee=convertir_en_temps(i['arrival_date_time'])))
pandas.DataFrame(rows)
# debug


# #### Question 2
# 
# Vous êtes un peu pressé et vous avez peur de vous tromper en arrivant à la gare car d'autres TGV partent à peu près en même temps (à partir de 19h00) de la gare de Lyon. Si vous demandez à l'API, combien de résultats vous donne-t-elle ?

# Soit l'API a changé, soit...

# In[16]:


requests.get(
    ('https://api.sncf.com/v1/coverage/sncf/stop_areas/'
     'stop_area:SNCF:87686006/departures'), auth=(token_auth, '')).json()


# In[17]:


### les trains qui partent autour de 19h00
departs_paris = requests.get(
    ('https://api.sncf.com/v1/coverage/sncf/stop_areas/stop_area:SNCF:'
     '87686006/departures?from_datetime={}').format(
        date_depart), auth=(token_auth, '')).json()

# Nombre de trains que l'API renvoie à partir de cet horaire-là
#print(len(departs_paris['departures']))
departs_paris


# In[18]:


from jyquickhelper import JSONJS
JSONJS(departs_paris)


# - Quels sont les horaires de départ de ces trains ? 

# In[19]:


for i in range(len(departs_paris['departures'])) :
    print(departs_paris['departures'][i]['stop_date_time']['departure_date_time'])


# - Parmi ces trains, combien de trains ont pour destination finale Lyon et qui partent le 17 novembre ?

# In[20]:


nombre_trains_pour_lyon = 0

for depart in departs_paris['departures'] :     
    if "Lyon" in depart['display_informations']['direction']:
        if (convertir_en_temps(depart['stop_date_time']['arrival_date_time']) > 
                convertir_en_temps(date_depart) and
                convertir_en_temps(depart['stop_date_time']['arrival_date_time']) < 
                datetime(2022,12,31,0,0,0)):
            nombre_trains_pour_lyon += 1
            print("le prochain départ pour Lyon sera le",
                  convertir_en_temps(depart['stop_date_time']['arrival_date_time']))
        
print("Il y a" , nombre_trains_pour_lyon, "train(s) pour Lyon dans les trains proposés", 
      "par l'API qui partent encore le 17 novembre")


# ### C'est quand qu'on va où ?
# 
# En fait, vous n'êtes plus très sûr de vouloir aller à Lyon. Mais bon maintenant vous êtes Gare de Lyon et il est 18h00. 

# #### Question 3
# 
# - Combien de tgv partent entre 18h00 et 20h00 ?
# - Lequel arrive le plus tôt à sa destination finale ?

# In[21]:


# on crée deux fonctions : 

def trouver_destination_tgv(origine, datetime) : 
    '''Permet d avoir les 10 prochains départs d une gare donnée '''
    return requests.get('https://api.sncf.com/v1/coverage/sncf/stop_areas/{}/' \
                        'departures?from_datetime={}'.format(origine, datetime) ,
                        auth=(token_auth, '')).json()

def trouver_trajet_dispo_max_heure(gare_depart, date_heure_depart, date_heure_max) : 
    '''Permet d avoir toutes les informations sur des trajets partant 
    d une gare entre une date X et une date Y'''
    
    destinations = []
    
    # on interroge l'API tant qu'il renvoie des information
    # sur les trains partant de Gare de lyon 

    while convertir_en_temps(date_heure_depart) < convertir_en_temps(date_heure_max) :
    # on prend toutes les destinations qui partent à partir d'une certaine heure
        destinations = destinations + trouver_destination_tgv(
            gare_depart, date_heure_depart)['departures']
        
        nombre_resultats = trouver_destination_tgv(
            gare_depart, date_heure_depart)['pagination']['items_on_page']
        
        if nombre_resultats <= 0:
            break

        # on trouve l'heure max de la première série de 10 solutions que l'application renvoie
        # on remplace l'heure qu'on cherche par celle là
        resultats = trouver_destination_tgv(gare_depart, date_heure_depart)
        date_heure_depart = resultats['departures'][nombre_resultats-1][
            'stop_date_time']['departure_date_time']

    return destinations


# In[22]:


# on trouve l'ensemble des trajets dont le départ est compris entre deux horaires
# informations sur le trajet qu'on choisit dans le futur
# l'API ne retourne pas de résultatq très loin dans le passé
now = datetime.now()
if now.hour < 6:
    # pas trop tôt
    now += timedelta(hours=4)
dt = now + timedelta(14)  # dans deux semaines

date_heure = convertir_en_chaine(dt)
max_date_heure = convertir_en_chaine(dt + timedelta(hours=24))
print("entre", date_heure, "et", max_date_heure)

gare_initiale = 'stop_area:SNCF:87686006'

# on demande à avoir tous les trajets partant de gare de lyon
# entre deux heures précises

destinations_depuis_paris_max_20h = trouver_trajet_dispo_max_heure(
    gare_initiale, date_heure, max_date_heure)

# on veut supprimer ceux pour lesquels le départ est après 20h00

dictionnaire_destinations = {}

i = 0

for depart in destinations_depuis_paris_max_20h :    
    print(depart['display_informations']['direction'],
          depart['stop_date_time']['departure_date_time'])
    if (convertir_en_temps(depart['stop_date_time']['departure_date_time']) < 
            convertir_en_temps(max_date_heure)): 
        i += 1
        dictionnaire_destinations[i] = depart 
    
print("Je peux prendre", len(dictionnaire_destinations.keys()), 
      "trains qui partent entre 18h et 20h de Gare de Lyon")


# In[23]:


dictionnaire_destinations


# ### Et les correspondances ? 

# #### Question 4
# 
# - On va essayer de voir jusqu'où on peut aller, en prenant des trains au départ de la Gare de Lyon : 
#     - Quelles sont toutes les gares atteignables en partant le 17 novembre, sans faire de changement et sans partir après minuit ?
#     - Si on prend un de ces trains, jusqu'où peut-on aller, avec une correspondance, sans partir après 8h le lendemain matin ?

# In[24]:


# on va trouver toutes les gares qui sont sur les trajets des 
# trains retenus donc atteignables sans correspondance

def trouver_toutes_les_gares_du_trajet(gare_depart, gare_arrivee_finale, horaire_depart) :
    return requests.get('https://api.sncf.com/v1/coverage/sncf/journeys?from={}&to={}' \
                        '&datetime={}'.format(gare_depart,gare_arrivee_finale,horaire_depart), \
                        auth=(token_auth, '')).json()


# In[25]:


# Exemple pour la première gare de la liste

if len(dictionnaire_destinations) > 1:
    gare_depart = dictionnaire_destinations[1]['stop_point']['id']
    gare_arrivee = dictionnaire_destinations[1]['route']['direction']['id']
    horaire_train = dictionnaire_destinations[1]['stop_date_time']['arrival_date_time']

    trajet_recherche = trouver_toutes_les_gares_du_trajet(gare_depart,gare_arrivee,horaire_train)
    session = trajet_recherche['journeys'][0]['sections'][0]
    if "stop_date_times" in session:
        for i in session['stop_date_times']:
            print(i['stop_point']['name'])


# In[26]:


# on fait la liste des gares où on
# peut aller sans correspondance

liste_gares_direct = []

for x in dictionnaire_destinations.keys():
    # on prend les deux gares départ + finale
    gare_depart = dictionnaire_destinations[x]['stop_point']['id']
    gare_arrivee = dictionnaire_destinations[x]['route']['direction']['id']
    horaire_train = dictionnaire_destinations[x]['stop_date_time']['arrival_date_time']
    
    # on appelle la fonction définie précédemment
    trajet_recherche = trouver_toutes_les_gares_du_trajet(gare_depart,gare_arrivee,horaire_train)
    if 'error' in trajet_recherche:
        continue
    session = trajet_recherche['journeys'][0]['sections'][0]
    if "stop_date_times" in session:
        for i in session['stop_date_times']: 
            print(i['stop_point']['name'], i['arrival_date_time'])
            liste_gares_direct.append(i['stop_point']['name'])
        print("-------------")    
    
#### là on a la liste des gares atteignables sans correspondance    
liste_gares_direct = set(liste_gares_direct)


# #### Exemple : trouver toutes les correspondances possibles depuis le trajet entre les gares de Paris et de Perpignan

# In[27]:


# pour le premier trajet gare de la liste trouvée à l'étape précédente
# on va chercher toutes les connexions des gares possibles, entre le moment de l'arrivée 
# et 8 heures le lendemain matin

if len(dictionnaire_destinations) > 1:
    gare_depart = dictionnaire_destinations[1]['stop_point']['id']
    gare_arrivee = dictionnaire_destinations[1]['route']['direction']['id']
    horaire_train = dictionnaire_destinations[1]['stop_date_time']['arrival_date_time']

    horaire_max = convertir_en_chaine(dt + timedelta(hours=8))
    print("horaire_max", horaire_max)
else:
    horaire_train = None


# In[28]:


trajet_recherche['journeys'][0]['sections'][0]


# #### Question 5
# - Représenter toutes les gares atteignables avec un graphique de type scatter. Distinguer les gares atteintes en un seul trajet et celles atteintes avec une correspondance.

# In[29]:


# on va les représenter grâce à la base des latitude / longitude

import pandas as pd
import matplotlib.pyplot as plt
import matplotlib as mpl
from matplotlib.lines import Line2D 

mpl.rcParams['axes.facecolor'] = "whitesmoke"

palette = plt.cm.spring 

liste_couleurs = [palette(0), palette(0.5), palette(0.8)]

data_all = pd.read_csv("./ensemble_gares.csv", encoding = "ISO-8859-1")

connexions = []
lat = []
lon = []
labels = []

dict_lat = data_all.set_index('name')['lat'].to_dict()
dict_lon = data_all.set_index('name')['lon'].to_dict()
#dict_lab = data_all.set_index('name')['name'].str.replace("gare de","").to_dict()


for gare in dict_type_gares: 
    if gare not in dict_lat:
        print("'{0}' pas trouvé dans dict_lat (problème d'accents?)".format(gare))
        continue
    if gare not in dict_lon:
        print("'{0}' pas trouvé dans dict_lon (problème d'accents?)".format(gare))
        continue
    lat.append(dict_lat[gare])                              
    lon.append(dict_lon[gare])
    labels.append(gare)


# In[30]:


get_ipython().run_line_magic('matplotlib', 'inline')


# In[31]:


### La carte 
###################################################################################################

def liste_unique(liste) : 
    unicite = [] 
    for x in liste : 
        if x in unicite :
            pass
        else :
            unicite.append(x)
    return unicite

lab_un = liste_unique(labels)
lat_un = liste_unique(lat)
lon_un = liste_unique(lon)

fig = plt.figure(figsize=(12,10))

for label, x, y in set(zip(labels, lon, lat)) :
    if dict_type_gares[label] == "direct" :  
        plt.annotate(label, xy = (x - 0.05, y - 0.05), horizontalalignment = 'right', size = 13)
    else :
        plt.annotate(label, xy = (x + 0.05, y + 0.05), horizontalalignment = 'left', size = 13)

colors = []     
for x in lab_un : 
    if dict_type_gares[x] == "depart" : 
        colors.append(liste_couleurs[0])
    if dict_type_gares[x] == "direct" :
        colors.append(liste_couleurs[1])
    if dict_type_gares[x] == "correspondance" : 
        colors.append(liste_couleurs[2])
                
        
plt.scatter(x = lon_un , y = lat_un, marker = "o", c = colors, s = 100, alpha = 0.5)

#### Legende

circ1 = Line2D([0], [0], linestyle="none", marker="o", alpha=0.5, markersize=10, markerfacecolor = liste_couleurs[0])
circ2 = Line2D([0], [0], linestyle="none", marker="o", alpha=0.5, markersize=10, markerfacecolor = liste_couleurs[1])
circ3 = Line2D([0], [0], linestyle="none", marker="o", alpha=0.5, markersize=10, markerfacecolor = liste_couleurs[2])

legende = plt.legend((circ1, circ2, circ3), ("Gare de départ", "Direct depuis Gare de Lyon le soir du 17 novembre", 
                                             "Avec une correspondance depuis une gare directe"), numpoints=1, loc="best")

legende.get_frame().set_facecolor('white')

plt.title("Gares atteignables avant minuit depuis la Gare de Lyon", size = 20);


# In[32]: